package com.jjt.order.service.impl;

import cn.hutool.json.JSONArray;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.jjt.common.constant.CacheConstants;
import com.jjt.common.core.redis.RedisCache;
import com.jjt.common.core.text.Convert;
import com.jjt.common.utils.DateUtils;
import com.jjt.common.utils.StringUtils;
import com.jjt.nccloud.utils.NccloudUtils;
import com.jjt.order.domain.*;
import com.jjt.order.mapper.TwinOrderMapper;
import com.jjt.order.service.ITwinBomService;
import com.jjt.order.service.ITwinOrderDetailService;
import com.jjt.order.service.ITwinOrderService;
import com.jjt.utils.MssqlUtil;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

/**
 * 销售订单信息Service业务层处理
 *
 * @author wukai
 * @date 2025-01-16
 */
@Service
public class TwinOrderServiceImpl implements ITwinOrderService {
    @Resource
    private TwinOrderMapper twinOrderMapper;
    @Resource
    private ITwinBomService bomService;
    @Resource
    private ITwinOrderDetailService detailService;
    @Resource
    private RedisCache redisCache;

    /**
     * 查询销售订单信息
     *
     * @param oderId 销售订单信息主键
     * @return 销售订单信息
     */
    @Override
    public TwinOrder selectTwinOrderByOderId(Long oderId) {
        return twinOrderMapper.selectTwinOrderByOderId(oderId);
    }

    /**
     * 查询销售订单信息列表
     *
     * @param twinOrder 销售订单信息
     * @return 销售订单信息
     */
    @Override
    public List<TwinOrder> selectTwinOrderList(TwinOrder twinOrder) {
        return twinOrderMapper.selectTwinOrderList(twinOrder);
    }

    /**
     * 新增销售订单信息
     *
     * @param twinOrder 销售订单信息
     * @return 结果
     */
    @Override
    public int insertTwinOrder(TwinOrder twinOrder) {
        int rows = twinOrderMapper.insertTwinOrder(twinOrder);
        insertTwinOrderDetail(twinOrder);
        return rows;
    }

    /**
     * 修改销售订单信息
     *
     * @param twinOrder 销售订单信息
     * @return 结果
     */
    @Override
    public int updateTwinOrder(TwinOrder twinOrder) {
        twinOrderMapper.deleteTwinOrderDetailByOderId(twinOrder.getOderId());
        insertTwinOrderDetail(twinOrder);
        return twinOrderMapper.updateTwinOrder(twinOrder);
    }

    /**
     * 批量删除销售订单信息
     *
     * @param oderIds 需要删除的销售订单信息主键
     * @return 结果
     */
    @Override
    public int deleteTwinOrderByOderIds(String oderIds) {
        twinOrderMapper.deleteTwinOrderDetailByOderIds(Convert.toStrArray(oderIds));
        return twinOrderMapper.deleteTwinOrderByOderIds(Convert.toStrArray(oderIds));
    }

    /**
     * 删除销售订单信息信息
     *
     * @param oderId 销售订单信息主键
     * @return 结果
     */
    @Override
    public int deleteTwinOrderByOderId(Long oderId) {
        return twinOrderMapper.deleteTwinOrderByOderId(oderId);
    }

    /**
     * 按时间获取订单入库
     *
     * @param localDate 时间
     */
    @Override
    public void getOrder(LocalDate localDate) {
        String req = "{\"pk_org\":\"100101\",\"dbilldate\":\"%s~%s\"}";
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
        LocalDateTime start = LocalDateTime.of(localDate, LocalTime.MIN);
        LocalDateTime end = start.plusDays(1);
        req = String.format(req, start.format(formatter), end.format(formatter));
        try {
            String jsonStr = NccloudUtils.getData(req);
            JSONObject jsonObject = JSONUtil.parseObj(jsonStr);
            if (!jsonObject.getBool("success")) {
                return;
            }

            List<TwinBom> bomList = bomService.selectTwinBomList(new TwinBom());
            Map<String, TwinBom> bomMap = bomList.stream().collect(Collectors.toMap(TwinBom::getBomCode, o -> o));

            JSONArray jsonArray = jsonObject.getJSONArray("data");
            // 输出 JsonArray 内容
            for (int i = 0; i < jsonArray.size(); i++) {
                TwinOrder twinOrder = new TwinOrder();
                JSONObject obj = jsonArray.getJSONObject(i);
                JSONObject order = obj.getJSONObject("so.so_saleorder");
                //客户号
                twinOrder.setCustomerCode(order.getByPath("ccustomerid.code", String.class));
                //客户名称
                twinOrder.setCustomerName(order.getByPath("ccustomerid.name", String.class));
                twinOrder.setOrderCode(order.getByPath("vbillcode", String.class));
                twinOrder.setOrderDate(DateUtils.parseDate(order.getByPath("dmakedate", String.class)));

                List<TwinOrderDetail> details = new ArrayList<>();
                JSONArray childOrder = obj.getJSONArray("so.so_saleorder_b");
                for (int j = 0; j < childOrder.size(); j++) {
                    JSONObject co = childOrder.getJSONObject(j);
                    TwinOrderDetail detail = new TwinOrderDetail();
                    String bomCode = co.getByPath("cmaterialvid.code", String.class);
                    detail.setUnit(co.getByPath("castunitid.name", String.class));
                    detail.setBomCode(bomCode);
                    detail.setBomName(co.getByPath("cmaterialvid.name", String.class));
                    detail.setColor(co.getByPath("vfree1.name", String.class));
                    detail.setPrice(co.getByPath("nqttaxnetprice", BigDecimal.class));
                    detail.setTotalPrice(co.getByPath("ntaxmny", BigDecimal.class));
                    detail.setNum(co.getByPath("nastnum", BigDecimal.class));
                    detail.setWeight(co.getByPath("nnum", BigDecimal.class));
                    TwinBom bom = bomMap.get(bomCode);
                    if (bom != null) {
                        detail.setBomWeight(bom.getBomWeight());
                        detail.setBomSpec(bom.getBomSpec());
                    }
                    details.add(detail);
                }
                twinOrder.setTwinOrderDetailList(details);
                insertTwinOrder(twinOrder);
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * 按时间获取订单信息
     *
     * @param start
     * @param end
     * @return
     */
    @Override
    public List<TwinOrder> selectTwinOrderListByDate(String start, String end) {
        TwinOrder order = new TwinOrder();
        Map<String, Object> params = new HashMap<>(16);
        params.put("start", start);
        params.put("end", end);
        order.setParams(params);
        List<TwinOrder> list = selectTwinOrderList(order);
        list.forEach(to -> {
            TwinOrderDetail tod = new TwinOrderDetail();
            tod.setOderId(to.getOderId());
            to.setTwinOrderDetailList(detailService.selectTwinOrderDetailList(tod));
        });
        return list;
    }

    /**
     * 获取当前库存
     */
    @Override
    public void currStock() {
        //减少7个小时，班组统计时间
        LocalDate localDate = LocalDateTime.now().minusHours(7).toLocalDate();
        List<MesStock> list = MssqlUtil.mesStock(localDate);
        Map<String, Double> currStock = MssqlUtil.currStock();
        HashMap<String, Object> initMap = new HashMap<>(16);
        initMap.put("d-in", 0L);
        initMap.put("d-out", 0L);
        initMap.put("m-in", 0L);
        initMap.put("m-out", 0L);
        Map<String, Object> mapping = new HashMap<>(16);
        mapping.put("WH01", "ycl");
        mapping.put("WH03", "pt");
        mapping.put("WH04", "bpb");
        mapping.put("WH08", "cp");
        Map<String, Object> result = new HashMap<>(16);
        result.put("ycl", initMap.clone());
        result.put("cp", initMap.clone());
        result.put("bpb", initMap.clone());
        result.put("pt", initMap.clone());

        for (String key : currStock.keySet()) {
            HashMap<String, Object> map = (HashMap<String, Object>) result.get(key);
            map.put("curr", currStock.get(key));
        }

        //先按仓库分组
        Map<String, List<MesStock>> codeMap = list.stream().collect(Collectors.groupingBy(MesStock::getCode));

        codeMap.forEach((code, codeList) -> {
            //按时间分组
            Map<String, Object> map = (Map<String, Object>) result.get(mapping.get(code));
            AtomicLong dIn = new AtomicLong((Long) map.get("d-in"));
            AtomicLong dOut = new AtomicLong((Long) map.get("d-out"));
            AtomicLong mIn = new AtomicLong((Long) map.get("m-in"));
            AtomicLong mOut = new AtomicLong((Long) map.get("m-out"));

            List<MesStockCalc> calcList = new ArrayList<>();
            Map<LocalDate, List<MesStock>> dateMap = codeList.stream().collect(Collectors.groupingBy(MesStock::getDate));
            dateMap.forEach((date, dateList) -> {
                List<MesStock> inList = dateList.stream().filter(MesStock::getFlag).collect(Collectors.toList());
                List<MesStock> outList = dateList.stream().filter(o -> !o.getFlag()).collect(Collectors.toList());
                //统计入库
                long inNum = inList.stream().mapToLong(MesStock::getTotal).sum();
                //统计出库
                long outNum = outList.stream().mapToLong(MesStock::getTotal).sum();

                MesStockCalc calc = new MesStockCalc();
                calc.setCode(code);
                calc.setDate(date);
                calc.setInNum(inNum);
                calc.setOutNum(outNum);
                calcList.add(calc);
                if (localDate.isEqual(date)) {
                    dIn.addAndGet(inNum);
                    dOut.addAndGet(outNum);
                    map.put("in", inList);
                    map.put("out", outList);
                }
                if (localDate.getYear() == date.getYear() && localDate.getMonthValue() == date.getMonthValue()) {
                    mIn.addAndGet(inNum);
                    mOut.addAndGet(outNum);
                }
                map.put("d-in", dIn.longValue());
                map.put("d-out", dOut.longValue());
                map.put("m-in", mIn.longValue());
                map.put("m-out", mOut.longValue());
            });
            calcList.sort(Comparator.comparing(MesStockCalc::getDate));
            map.put("trend", calcList);
        });
        redisCache.setCacheObject(CacheConstants.VMS_STOCK, result);
    }

    /**
     * 新增销售订单明细信息信息
     *
     * @param twinOrder 销售订单信息对象
     */
    public void insertTwinOrderDetail(TwinOrder twinOrder) {
        List<TwinOrderDetail> twinOrderDetailList = twinOrder.getTwinOrderDetailList();
        Long oderId = twinOrder.getOderId();
        if (StringUtils.isNotNull(twinOrderDetailList)) {
            List<TwinOrderDetail> list = new ArrayList<TwinOrderDetail>();
            for (TwinOrderDetail twinOrderDetail : twinOrderDetailList) {
                twinOrderDetail.setOderId(oderId);
                list.add(twinOrderDetail);
            }
            if (list.size() > 0) {
                twinOrderMapper.batchTwinOrderDetail(list);
            }
        }
    }
}
